unit uart;
{$xdata}

interface

function lf : boolean;
procedure poll; usingany;
procedure rate ( baud : longint );

implementation
const
    rmask = $FF;
    tmask = $3F;
var
    t2mod : byte absolute $C9; volatile;
    rbuf : array [0..rmask] of byte xdata;
    tbuf : array [0..tmask] of byte xdata;
    rin, rout, tin, tout, tmp : byte;
    zlf : boolean;

function lf;
var
    i : byte;
begin
    i := rout;
    result := false;
    while i <> rin do begin
        if rbuf[i] = system.lf then begin
            result := true;
            break;
        end;
        i := (i + 1) and rmask;
    end;
end;

procedure poll;
begin
    if ri then begin
        ri := false;
        rbuf[rin] := sbuf;
        rin := (rin + 1) and rmask;
    end;
    if (tin <> tout) and ti then begin
        ti := false;
        sbuf := tbuf[tout];
        tout := (tout + 1) and tmask;
    end;
end;

procedure rate;
begin
    scon := 0;
    if baud > 6000 then begin
        t2mod := t2mod or {tmr_clk t1_clk} $A0;
        th1 := -byte( 1500000 div baud );
    end else if baud > 500 then begin
        t2mod := t2mod and (not {t1_clk} $20);
        th1 := -byte( 125000 div baud );
    end;
    tl1 := $FF;
    scon := {sm1 ren ti} $52;
end;

function read : byte;
begin
    if zlf then begin
        zlf := false;
        tmp := system.lf;
    end else if rin = rout then begin
        zlf := true;
        tmp := 0;
    end else begin
        tmp := rbuf[rout];
        rout := (rout + 1) and rmask;
    end;
    result := tmp;
end;

procedure write;
begin
    asm mov tmp, a end;
    tbuf[tin] := tmp;
    tin := (tin + 1) and tmask;
end;

begin
    assign( systemio, read, write );
    rin := 0;
    rout := 0;
    tin := 0;
    tout := 0;
    zlf := false;
    pcon := {smod} $80;
    rate( 9600 );
end.
