unit usb;
{$xdata}

interface
var 
    buf : array [1..3] of record
        r, t : array [0..63] of byte end xdata; volatile;
    len : array [1..3] of record
        r, t : byte end; volatile;
    ok : boolean;

procedure poll;
procedure _usb; interrupt $43; using 2;

implementation
const
    vid = $6666;
    pid = $5500;
    rev = $0123;
    device : array [1..9] of word = (
        $112, $110, 0, $4000, vid, pid, rev, $200, $103
        );
    config : array [1..60] of byte = (
        9, 2, 60,0, 1, 1, 0, $80, 100,
        9, 4, 0, 0, 6, $FF, 0, 0, 0,
        7, 5, $81, 2, 64,0, 0,
        7, 5,   1, 2, 64,0, 0,
        7, 5, $82, 2, 64,0, 0,
        7, 5,   2, 2, 64,0, 0,
        7, 5, $83, 2, 64,0, 0,
        7, 5,   3, 2, 64,0, 0
        );
    lang   : array [1..4]  of byte = ( 4, 3, 9,4 );
    prod   : array [1..12] of byte = ( 12, 3, 90,0, 97,0, 100,0, 105,0, 103,0 );
    serial : array [1..12] of byte = ( 12, 3, 49,0, 50,0, 51,0, 52,0, 53,0 );
var
    ie_ex      : byte absolute $E8; volatile;
    ie_usb     : boolean absolute ie_ex.2;
    ip_ex      : byte absolute $E9; volatile;
    udev_ctrl  : byte absolute $D1; volatile;
    uep_mod    : word absolute $EA; volatile;
    uep0_ctrl  : byte absolute $DC; volatile;
    uep0_dma   : word absolute $EC; volatile;
    uep0_t_len : byte absolute $DD; volatile;
    uep1_ctrl  : byte absolute $D2; volatile;
    uep1_dma   : word absolute $EE; volatile;
    uep1_t_len : byte absolute $D3; volatile;
    uep2_ctrl  : byte absolute $D4; volatile;
    uep2_dma   : word absolute $E4; volatile;
    uep2_t_len : byte absolute $D5; volatile;
    uep3_ctrl  : byte absolute $D6; volatile;
    uep3_dma   : word absolute $E6; volatile;
    uep3_t_len : byte absolute $D7; volatile;
    usb_ctrl   : byte absolute $E2; volatile;
    usb_dev_ad : byte absolute $E3; volatile;
    usb_int_en : byte absolute $E1; volatile;
    usb_int_fg : byte absolute $D8; volatile;
    uif_reset  : boolean absolute usb_int_fg.0;
    uif_trans  : boolean absolute usb_int_fg.1;
    usb_int_st : byte absolute $D9; volatile;
    usb_rx_len : byte absolute $DB; volatile;
    address : byte;
    buf0 : array [0..31] of word xdata; volatile;
    stage : ( setup, data, status );

procedure reset; usingany;
begin
    address := 0;
    len[1].r := 0; len[1].t := 0;
    len[2].r := 0; len[2].t := 0;
    len[3].r := 0; len[3].t := 0;
    ok := false;
    stage := setup;
    uep0_ctrl := {r-nak t-nak} $A;
    uep1_ctrl := {r-auto t-auto r-nak t-nak} $3A;
    uep2_ctrl := {r-auto t-auto r-nak t-nak} $3A;
    uep3_ctrl := {r-auto t-auto r-nak t-nak} $3A;
    usb_dev_ad := 0;
    uif_reset := false;
    uif_trans := false;
end;

procedure trans; usingany;
var
    len0 : byte;
begin
    case usb_int_st and $3F of
        {setup} $30 : begin
            len0 := byte( buf0[3] );
            stage := setup;
            case buf0[0] of
                {get descriptor} $680 : case buf0[1] of
                    {device} $100 : begin
                        move( device, buf0, sizeof( device ) );
                        stage := data;
                    end;
                    {configuration} $200 : begin
                        move( config, buf0, sizeof( config ) );
                        stage := data;
                    end;
                    {language} $300 : begin
                        move( lang, buf0, sizeof( lang ) );
                        stage := data;
                    end;
                    {product} $302 : begin
                        move( prod, buf0, sizeof( prod ) );
                        stage := data;
                    end;
                    {serial} $303 : begin
                        move( serial, buf0, sizeof( serial ) );
                        stage := data;
                    end;
                end;
                {get status} $80, $81, $82,
                {get configuration} $880,
                {get interface} $A81 : begin
                    buf0[0] := 0;
                    stage := data;
                end;
                {set address} $500 : begin
                    address := buf0[1];
                    stage := status;
                end;
                {clear future} $100, $101, $102,
                {set future}   $300, $301, $302,
                {set configuration} $900,
                {set interface} $B01 : begin
                    ok := true;
                    stage := status;
                end;
            end;
            case stage of
                data : begin
                    if len0 > 63 then len0 := 63;
                    uep0_t_len := len0;
                    uep0_ctrl := {r-data1 t-data1 r-nak t-ack} $C8;
                end;
                status : begin
                    uep0_t_len := 0;
                    uep0_ctrl := {r-data1 t-data1 r-nak t-ack} $C8;
                end;
                else uep0_ctrl := {r-stall t-stall} $F;
            end;
        end;
        {out0} 0 : case stage of
            status : begin
                uep0_ctrl := {r-nak t-nak} $A;
                stage := setup;
            end;
            else uep0_ctrl := {r-stall t-stall} $F;
        end;
        {in0} $20 : case stage of
            data : begin
                uep0_ctrl := {r-data1 t-data1 r-ack t-nak} $C2;
                stage := status;
            end;
            status : begin
                usb_dev_ad := address;
                uep0_ctrl := {r-nak t-nak} $A;
                stage := setup;
            end;
            else uep0_ctrl := {r-stall t-stall} $F;
        end;
        {out1} 1 : begin
            {r-nak} uep1_ctrl := uep1_ctrl or 8;
            len[1].r := usb_rx_len;
        end;
        {in1} $21 : begin
            {t-nak} uep1_ctrl := uep1_ctrl or 2;
            len[1].t := 0;
        end;
        {out2} 2 : begin
            {r-nak} uep2_ctrl := uep2_ctrl or 8;
            len[2].r := usb_rx_len;
        end;
        {in2} $22 : begin
            {t-nak} uep2_ctrl := uep2_ctrl or 2;
            len[2].t := 0;
        end;
        {out3} 3 : begin
            {r-nak} uep3_ctrl := uep3_ctrl or 8;
            len[3].r := usb_rx_len;
        end;
        {in3} $23 : begin
            {t-nak} uep3_ctrl := uep3_ctrl or 2;
            len[3].t := 0;
        end;
    end;
    uif_trans := false;
end;

procedure poll;
begin
    if ((uep1_ctrl and $C) <> 0) and (len[1].r = 0) then begin
        {r-ack} uep1_ctrl := uep1_ctrl and (not $C);
    end;
    if ((uep1_ctrl and 3) <> 0) and (len[1].t <> 0) then begin
        if len[1].t > 64 then len[1].t := 64;
        uep1_t_len := len[1].t;
        {t-ack} uep1_ctrl := uep1_ctrl and (not 3);
    end;
    if ((uep2_ctrl and $C) <> 0) and (len[2].r = 0) then begin
        {r-ack} uep2_ctrl := uep2_ctrl and (not $C);
    end;
    if ((uep2_ctrl and 3) <> 0) and (len[2].t <> 0) then begin
        if len[2].t > 64 then len[2].t := 64;
        uep2_t_len := len[2].t;
        {t-ack} uep2_ctrl := uep2_ctrl and (not 3);
    end;
    if ((uep3_ctrl and $C) <> 0) and (len[3].r = 0) then begin
        {r-ack} uep3_ctrl := uep3_ctrl and (not $C);
    end;
    if ((uep3_ctrl and 3) <> 0) and (len[3].t <> 0) then begin
        if len[3].t > 64 then len[3].t := 64;
        uep3_t_len := len[3].t;
        {t-ack} uep3_ctrl := uep3_ctrl and (not 3);
    end;
    if not ea then begin
        if uif_reset then reset;
        if uif_trans then trans;
    end;
end;

procedure _usb;
begin
    if uif_reset then reset;
    if uif_trans then trans;
end;

begin
    ie_usb := false;
    usb_ctrl := 0;
    uep_mod := {3-rt 2-rt 1-rt 4-} $CCC0;
    uep0_dma := word( addr( buf0 ) );
    uep1_dma := word( addr( buf[1] ) );
    uep2_dma := word( addr( buf[2] ) );
    uep3_dma := word( addr( buf[3] ) );
    reset;
    udev_ctrl := {pd_dis} $80;
    usb_ctrl := {dev_pu_en int_busy dma_en} $29;
    udev_ctrl := {pd_dis port_en} $81;
    usb_int_en := {trans reset} 3;
    ip_ex := {ip_usb} 4;
    ie_usb := true;
end.
